/***************************************************************************
 *
 * Copyright (C) 2007  SMSC
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 ***************************************************************************
 * File: bin_merger.c
 *
 * Revision History
 *  Rev 1.0.2	5/01/2007 WH
 *  	- fix the bug that write_outfile() doesn't return 0 when success
 *
 *  Rev 1.0.1	3/30/2007 WH
 *  	- update outfile creation file attribute problem
 *  	- Add "Error!" string at the begin of message
 *
 *  Rev 1.0.0	3/29/2007 WH
 *  	- first release
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#undef	DEBUG

#define	VERSION					"1.0.2"

#define	LINE_BUF_SIZE			(1024)
#define	DEFAULT_OUTFILENAME		"flash_img.bin"
#define	DEFAULT_FLASH_SIZE		(0x01000000)			// 16Mbytes as default flash size
#define DEFAULT_FILLBYTE		(0xFF)
#define	FILE_BUFFER_SIZE		(1024*1024)

#ifdef	DEBUG
#define	DPRINT(s, arg...)	do { fprintf(stdout, "[%s() #%d] " s, __FUNCTION__, __LINE__, ##arg); } while(0)
#else
#define	DPRINT(s, arg...)	while (0)
#endif

typedef struct  file_config_t	*p_file_config;

struct	file_config_t	{
	char	filename[LINE_BUF_SIZE];
	int		filesize;
	ulong	offset;
	p_file_config	prev;
	p_file_config	next;
} file_config;

struct	global_config_t {
	char	outfilename[LINE_BUF_SIZE];
	ulong	maxsize;
	unsigned char	fillbyte;
	p_file_config	fc;
}	global_config;

void usage(char *filename)
{
	fprintf(stdout, "Usage: %s config_file\n", filename);
}

static void print_fc_list()
{
	p_file_config fc;

	if (global_config.fc == NULL)
	{
		fprintf(stdout, "file config list is empty\n");
		return;
	}

	fc = global_config.fc;

	while (fc != NULL) {
		fprintf(stdout, "filename = %s\n", fc->filename);
		fprintf(stdout, "filesize = %ld\n", fc->filesize);
		fprintf(stdout, "offset = 0x%08lx\n", fc->offset);
		fc = fc->next;
	}
}

FILE *open_config_file(char *filename)
{
	FILE *fd;

	DPRINT("open %s\n", filename);

	fd = fopen(filename, "r");

	return fd;
}

void free_fc_list()
{
	p_file_config	fc1, fc2;

	DPRINT("free file config list\n");

	fc1 = fc2 = global_config.fc;
	while (fc2 != NULL)
	{
		fc2 = fc1->next;
		free(fc1);
		fc1 = fc2;
	}
}

int close_config_file(FILE *fd)
{
	if (fd != NULL)
	{
		free_fc_list();
		fclose(fd);
		return 0;
	}
	else
	{
		DPRINT("fd = NULL\n");
		return -1;
	}
}

int parse_config_command(char *line, char *item, char *value)
{
	char	*buf[LINE_BUF_SIZE];
	char	*data;

	data = strtok_r(line, "=", buf);
	if (data == NULL)
	{
		fprintf(stderr, "Error! syntax error at %s\n", line);
		return -1;
	}
	strcpy(item, data);

	data = strtok_r(NULL, "=", buf);
	if (data == NULL)
	{
		fprintf(stderr, "Error! syntax error at %s\n", line);
		return -1;
	}
	strcpy(value, data);

	if (strtok_r(NULL, "=", buf) != NULL)
	{
		fprintf(stderr, "Error! syntax error at %s\n", line);
		return -1;
	}

	return 0;
}

int convert_item_value(p_file_config fc, char *item, char *value)
{
	struct stat filestat;

	if (fc == NULL)
	{
		fprintf(stderr, "Error! memory allocation error\n");
		return -1;
	}

	if (!strcmp(item, "filename"))
	{
		strcpy(fc->filename, value);
		DPRINT("filename = %s\n", fc->filename);
		if (stat(fc->filename, &filestat) < 0)
		{
			fprintf(stderr, "Error! can not get stat of file(%s)\n", fc->filename);
			return -2;
		}
		fc->filesize = filestat.st_size;
		DPRINT("filesize = %dbytes\n", fc->filesize);
	}
	else if (!strcmp(item, "offset"))
	{
		fc->offset = strtoul(value, NULL, 0);
		DPRINT("offset = 0x%08lx\n", fc->offset);
	}
	else if (!strcmp(item, "outfile"))
	{
		strcpy(global_config.outfilename, value);
		DPRINT("outfile = %s\n", global_config.outfilename);
	}
	else if (!strcmp(item, "maxsize"))
	{
		global_config.maxsize = strtoul(value, NULL, 0);
		DPRINT("maxsize = 0x%08lx\n", global_config.maxsize);
	}
	else if (!strcmp(item, "fillbyte"))
	{
		global_config.fillbyte = (unsigned char)strtoul(value, NULL, 0);
		DPRINT("fillbyte = 0x%02x\n", global_config.fillbyte);
	}
}

int add_fc(p_file_config fc)
{
	if (global_config.fc == NULL)
	{
		DPRINT("insert into first entry of list\n");
		// first item
		fc->prev = fc->next = NULL;
		global_config.fc = fc;
	}
	else
	{
		// add into list
		p_file_config	tmp;

		// find where to insert based on fc->offset
		tmp = global_config.fc;
		do {
			if (tmp->offset == fc->offset)
			{
				fprintf(stderr, "Error! found duplicated offset of 0x%08lx\n", fc->offset);
				return -1;
			}
			else if (tmp->offset > fc->offset)
			{
				// insert before tmp
				DPRINT("found offset 0x%08lx, insert offset 0x%08lx in front of it\n", 
						tmp->offset, fc->offset);
				fc->prev = tmp->prev;
				if (fc->prev)
				{
					fc->prev->next = fc;
				}
				else
					global_config.fc = fc;
				tmp->prev = fc;
				fc->next = tmp;
				return 0;
			}
			if (tmp->next == NULL)
				break;
			else
				tmp = tmp->next;
		} while (1);

		DPRINT("insert offset 0x%08lx at the end of list\n", fc->offset);
		// append at the end of list
		tmp->next = fc;
		fc->prev = tmp;
		fc->next = NULL;
	}

	return 0;
}

int parse_config_line(char *line)
{
	char	*buf[LINE_BUF_SIZE];
	char	*single_config;
	char	item[LINE_BUF_SIZE], value[LINE_BUF_SIZE];	
	p_file_config	fc = NULL;

	single_config = strtok_r(line, ", ", buf);
	if (single_config)
	{
		fc = (p_file_config)malloc(sizeof(file_config));
		memset(fc, 0, sizeof(*fc));
	}

	while (single_config != NULL)
	{
		if (parse_config_command(single_config, item, value) < 0)
		{
			fprintf(stderr, "Error! error at %s\n", single_config);
			free(fc);
			return -1;
		}
		else
		{
			convert_item_value(fc, item, value);
		}
		single_config = strtok_r(NULL, ", ", buf);
	}

	if ((strlen(fc->filename) == 0) || (fc->filesize == 0))
	{
		// this is line for "outfilename.."
		free(fc);
	}
	else
	{
		// add fc to global_config
		add_fc(fc);
	}

	return 0;
}

int verify_address_overwrap()
{
	p_file_config	fc;

	fc = global_config.fc;
	if (fc == NULL)
	{
		fprintf(stderr, "Error! no file list to check\n");
		return -1;
	}

	while (fc->next != NULL)
	{
		if (fc->offset >= global_config.maxsize)
		{
			fprintf(stderr, "Error! total size is bigger than maxsize (0x%08lx)\n", global_config.maxsize);
			return -2;
		}
		if ((fc->offset + fc->filesize) > fc->next->offset)
		{
			fprintf(stderr, "Error! %s is overwrap to next file\n", fc->filename);
			return -3;
		}
		fc = fc->next;
	}

	// now last one..
	if ((fc->offset + fc->filesize) > global_config.maxsize)
	{
		fprintf(stderr, "Error! %s is overwrap to next file\n", fc->filename);
		return -3;
	}

	return 0;
}

int parse_config_file(FILE *fd)
{
	char	buf[LINE_BUF_SIZE];

	fprintf(stdout, "parse config file\n");

	// read configuration and fill global_config structure
	while (fgets(buf, LINE_BUF_SIZE, fd) != NULL)
	{
		if (parse_config_line(buf) < 0)
			return -1;
	}

	// verify address overwrap
	if (verify_address_overwrap() < 0)
	{
		return -2;
	}
	
	return 0;
}

void set_default_config()
{
	memset(&global_config, 0, sizeof(global_config));

	strcpy(global_config.outfilename, DEFAULT_OUTFILENAME);
	global_config.maxsize = DEFAULT_FLASH_SIZE;
	global_config.fillbyte = DEFAULT_FILLBYTE;
	global_config.fc = NULL;
}

int write_fillbyte(int fd, ulong start, ulong end)
{
	ulong	i;

	// fill
	for (i=start;i<end;i++)
	{
		write(fd, &global_config.fillbyte, 1);
	}
	DPRINT("fill %10d bytes\n", end - start);

	return 0;
}

int write_each_file_to_outfile(int fo, char *filename, int filesize)
{
	int	len;
	int	total_len;
	int	fi;
	unsigned char	buf[FILE_BUFFER_SIZE];

	fi = open(filename, O_RDONLY);
	if (fi < 0)
	{
		fprintf(stderr, "Error! fail to open file (%s)\n", filename);
		return -1;
	}

	total_len = 0;
	while ((len = read(fi, buf, FILE_BUFFER_SIZE)) != 0)
	{
		int	ret;

		ret = write(fo, buf, len);
		if (ret < 0)
		{
			fprintf(stderr, "Error! error to write to outfile (errno=%d)\n", errno);
			return -2;
		}
		total_len += ret;
	}

	DPRINT("written %10d bytes\n", total_len);

	close(fi);
	return 0;
}

int write_outfile()
{
	int	fd;
	ulong	saddr;
	p_file_config	fc;

	fd = open(global_config.outfilename, O_RDWR | O_CREAT | O_TRUNC, S_IRWXU);
	if (fd < 0)
	{
		fprintf(stderr, "Error! fail to create output file (%s)\n", global_config.outfilename);
		return -1;
	}

	fc = global_config.fc;
	if (fc == NULL)
	{
		fprintf(stderr, "Error! invalid config file list\n");
		close(fd);
		return -2;
	}

	fprintf(stdout, "creating %s\n", global_config.outfilename);
	saddr = 0;
	while (fc != NULL)
	{
		write_fillbyte(fd, saddr, fc->offset);
		write_each_file_to_outfile(fd, fc->filename, fc->filesize);
		saddr = fc->offset + fc->filesize;
		fc = fc->next;
	}
	// write fillbyte upto maxsize
	write_fillbyte(fd, saddr, global_config.maxsize);

	close(fd);
	return 0;
}

int main(int argc, char *argv[])
{
	int	ret;
	FILE *cfg_fd;

	fprintf(stdout, "binary merger ver %s\n", VERSION);

	if (argc != 2)
	{
		usage(argv[0]);
		exit(1);
	}

	// open config file
	cfg_fd = open_config_file(argv[1]);
	if (cfg_fd == NULL)
	{
		fprintf(stderr, "Error! failed to open config file %s. (errno = %d)\n", argv[1], errno);
		exit(2);
	}

	// set default global configuration
	set_default_config();

	// parse config file
	if (parse_config_file(cfg_fd) < 0)
	{
		close_config_file(cfg_fd);
		exit(3);
	}

	// write to out file
	if (write_outfile() < 0)
	{
		close_config_file(cfg_fd);
		exit(4);
	}
	
	fprintf(stdout, "done\n");

	// close config file and free resources
	close_config_file(cfg_fd);
	exit(0);
}

